@@ -43,53 +43,68 @@ class AsynCommandGenerator:
     def __del__(self): self.uncfgCmdGen()
 
     def cfgCmdGen(self, authData, transportTarget):
-        if authData.securityName not in self.__knownAuths:
-            if isinstance(authData, CommunityData):
+        if isinstance(authData, CommunityData):
+            if authData.communityIndex not in self.__knownAuths:
                 config.addV1System(
                     self.snmpEngine,
-                    authData.securityName,
+                    authData.communityIndex,
                     authData.communityName,
                     authData.contextEngineId,
                     authData.contextName,
-                    authData.tag
+                    authData.tag,
+                    authData.securityName
                 )
-            elif isinstance(authData, UsmUserData):
+                self.__knownAuths[authData.communityIndex] = authData
+        elif isinstance(authData, UsmUserData):
+            authDataKey = authData.userName, authData.securityEngineId
+            if authDataKey not in self.__knownAuths:
                 config.addV3User(
                     self.snmpEngine,
-                    authData.securityName,
+                    authData.userName,
                     authData.authProtocol, authData.authKey,
                     authData.privProtocol, authData.privKey,
-                    authData.contextEngineId
+                    authData.securityEngineId,
+                    securityName=authData.securityName
                 )
-            else:
-                raise error.PySnmpError('Unsupported authentication object')
+                self.__knownAuths[authDataKey] = authData
+        else:
+            raise error.PySnmpError('Unsupported authentication object')
 
-            self.__knownAuths[authData.securityName] = authData
-
-        k = authData.securityName, authData.securityLevel, authData.mpModel
-        if k in self.__knownParams:
-            paramsName = self.__knownParams[k]
+        paramsKey = authData.securityName, \
+                    authData.securityLevel, \
+                    authData.mpModel
+        if paramsKey in self.__knownParams:
+            paramsName, useCount = self.__knownParams[paramsKey]
+            self.__knownParams[paramsKey] = paramsName, useCount + 1
         else:
             paramsName = 'p%s' % nextID()
             config.addTargetParams(
                 self.snmpEngine, paramsName,
                 authData.securityName, authData.securityLevel, authData.mpModel
             )
-            self.__knownParams[k] = paramsName
+            self.__knownParams[paramsKey] = paramsName, 1
 
-        if transportTarget.transportDomain not in self.__knownTransports:
+        if transportTarget.transportDomain in self.__knownTransports:
+            transportTarget.verifyDispatcherCompatibility(self.snmpEngine)
+            transport, useCount = self.__knownTransports[transportTarget.transportDomain]
+            self.__knownTransports[transportTarget.transportDomain] = transport, useCount + 1
+        else:
             transport = transportTarget.openClientMode()
-            config.addSocketTransport(
+            config.addTransport(
                 self.snmpEngine,
                 transportTarget.transportDomain,
                 transport
             )
-            self.__knownTransports[transportTarget.transportDomain] = transport
+            self.__knownTransports[transportTarget.transportDomain] = transport, 1
 
-        k = paramsName, transportTarget, transportTarget.tagList
+        transportKey = ( paramsName,
+                         transportTarget.transportDomain,
+                         transportTarget.transportAddr,
+                         transportTarget.tagList )
 
-        if k in self.__knownTransportAddrs:
-            addrName = self.__knownTransportAddrs[k]
+        if transportKey in self.__knownTransportAddrs:
+            addrName, useCount = self.__knownTransportAddrs[transportKey]
+            self.__knownTransportAddrs[transportKey] = addrName, useCount + 1
         else:
             addrName = 'a%s' % nextID()
             config.addTargetAddr(
@@ -101,57 +116,115 @@ class AsynCommandGenerator:
                 transportTarget.retries,
                 transportTarget.tagList
             )
-            self.__knownTransportAddrs[k] = addrName
+            self.__knownTransportAddrs[transportKey] = addrName, 1
 
         return addrName, paramsName
 
-    def uncfgCmdGen(self):
-        for authData in self.__knownAuths.values():
+    def uncfgCmdGen(self, authData=None):
+        if authData:
             if isinstance(authData, CommunityData):
-                config.delV1System(
-                    self.snmpEngine,
-                    authData.securityName
-                    )
+                authDataKey = authData.communityIndex
             elif isinstance(authData, UsmUserData):
-                config.delV3User(
-                    self.snmpEngine, authData.securityName
-                    )
+                authDataKey = authData.userName, authData.securityEngineId
             else:
                 raise error.PySnmpError('Unsupported authentication object')
-        self.__knownAuths.clear()
+            if authDataKey in self.__knownAuths:
+                authDataKeys = ( authDataKey, )
+            else:
+                raise error.PySnmpError('Unknown authData %s' % (authData,))
+        else:
+            authDataKeys = list(self.__knownAuths.keys())
 
-        for paramsName in self.__knownParams.values():
-            config.delTargetParams(
-                self.snmpEngine, paramsName
-                )
-        self.__knownParams.clear()
-        
-        for transportDomain, transport in self.__knownTransports.items():
-            config.delSocketTransport(
-                self.snmpEngine, transportDomain
-                )
-            transport.closeTransport()
-        self.__knownTransports.clear()
+        addrNames, paramsNames = set(), set()
 
-        for addrName in self.__knownTransportAddrs.values():
-            config.delTargetAddr(
-                self.snmpEngine, addrName
+        for authDataKey in authDataKeys:
+            authDataX = self.__knownAuths[authDataKey] 
+            del self.__knownAuths[authDataKey]
+            if isinstance(authDataX, CommunityData):
+                config.delV1System(
+                    self.snmpEngine,
+                    authDataX.communityIndex
                 )
-        self.__knownTransportAddrs.clear()
-                
+            elif isinstance(authDataX, UsmUserData):
+                config.delV3User(
+                    self.snmpEngine,
+                    authDataX.userName, 
+                    authDataX.securityEngineId
+                )
+            else:
+                raise error.PySnmpError('Unsupported authentication object')
+
+            paramsKey = authDataX.securityName, \
+                        authDataX.securityLevel, \
+                        authDataX.mpModel
+            if paramsKey in self.__knownParams:
+                paramsName, useCount = self.__knownParams[paramsKey]
+                useCount -= 1
+                if useCount:
+                    self.__knownParams[paramsKey] = paramsName, useCount
+                else:
+                    del self.__knownParams[paramsKey]
+                    config.delTargetParams(
+                        self.snmpEngine, paramsName
+                    )
+                    paramsNames.add(paramsName)
+            else:
+                raise error.PySnmpError('Unknown target %s' % (paramsKey,))
+
+            addrKeys = [ x for x in self.__knownTransportAddrs if x[0] == paramsName ]
+
+            for addrKey in addrKeys:
+                addrName, useCount = self.__knownTransportAddrs[addrKey]
+                useCount -= 1
+                if useCount:
+                    self.__knownTransportAddrs[addrKey] = addrName, useCount
+                else:
+                    config.delTargetAddr(self.snmpEngine, addrName)
+
+                    addrNames.add(addrKey)
+
+                    if addrKey[1] in self.__knownTransports:
+                        transport, useCount = self.__knownTransports[addrKey[1]]
+                        if useCount > 1:
+                            useCount -= 1
+                            self.__knownTransports[addrKey[1]] = transport, useCount
+                        else:
+                            config.delTransport(self.snmpEngine, addrKey[1])
+                            transport.closeTransport()
+                            del self.__knownTransports[addrKey[1]]
+
+        return addrNames, paramsNames
+
+    # compatibility stub
     def makeReadVarBinds(self, varNames):
-        varBinds = []
-        for varName in varNames:
+        return self.makeVarBinds(
+            [ (x, self._null) for x in varNames ], oidOnly=True
+        )
+
+    def makeVarBinds(self, varBinds, oidOnly=False):
+        __varBinds = []
+        for varName, varVal in varBinds:
             if isinstance(varName, MibVariable):
-                varName.resolveWithMib(
-                    self.mibViewController, oidOnly=True
-                )
+                if oidOnly or isinstance(varVal, base.AbstractSimpleAsn1Item):
+                    varName.resolveWithMib(self.mibViewController, oidOnly=True)
+                else:
+                    varName.resolveWithMib(self.mibViewController)
+                    varVal = varName.getMibNode().getSyntax().clone(varVal)
             elif isinstance(varName[0], tuple):  # legacy
                 varName = MibVariable(varName[0][0], varName[0][1], *varName[1:]).resolveWithMib(self.mibViewController)
+                if not oidOnly and \
+                        not isinstance(varVal, base.AbstractSimpleAsn1Item):
+                    varVal = varName.getMibNode().getSyntax().clone(varVal)
             else:
-                varName = MibVariable(varName).resolveWithMib(self.mibViewController, oidOnly=True)
-            varBinds.append((varName, self._null))
-        return varBinds
+                if oidOnly or isinstance(varVal, base.AbstractSimpleAsn1Item):
+                    varName = MibVariable(varName).resolveWithMib(self.mibViewController, oidOnly=True)
+                else:
+                    varName = MibVariable(varName).resolveWithMib(self.mibViewController)
+                    varVal = varName.getMibNode().getSyntax().clone(varVal)
+
+            __varBinds.append((varName, varVal))
+
+        return __varBinds
 
     def unmakeVarBinds(self, varBinds, lookupNames, lookupValues):
         if lookupNames or lookupValues:
@@ -174,7 +247,8 @@ class AsynCommandGenerator:
     # Async SNMP apps
 
     def getCmd(self, authData, transportTarget, varNames, cbInfo,
-               lookupNames=False, lookupValues=False):
+               lookupNames=False, lookupValues=False,
+               contextEngineId=None, contextName=null):
         def __cbFun(sendRequestHandle,
                     errorIndication, errorStatus, errorIndex,
                     varBinds, cbCtx):
@@ -188,23 +262,29 @@ class AsynCommandGenerator:
                 cbCtx
             )
 
+        # for backward compatibility
+        if contextName is null and authData.contextName:
+            contextName = authData.contextName
+        
         (cbFun, cbCtx) = cbInfo
         addrName, paramsName = self.cfgCmdGen(
             authData, transportTarget
         )
+
         return cmdgen.GetCommandGenerator().sendReq(
             self.snmpEngine,
             addrName,
             self.makeReadVarBinds(varNames),
             __cbFun,
             (lookupNames, lookupValues, cbFun, cbCtx),
-            authData.contextEngineId, authData.contextName
+            contextEngineId, contextName
         )
     
     asyncGetCmd = getCmd
     
     def setCmd(self, authData, transportTarget, varBinds, cbInfo,
-               lookupNames=False, lookupValues=False):
+               lookupNames=False, lookupValues=False,
+               contextEngineId=None, contextName=null):
         def __cbFun(sendRequestHandle,
                     errorIndication, errorStatus, errorIndex,
                     varBinds, cbCtx):
@@ -218,42 +298,30 @@ class AsynCommandGenerator:
                 cbCtx
             )
 
+        # for backward compatibility
+        if contextName is null and authData.contextName:
+            contextName = authData.contextName
+        
         (cbFun, cbCtx) = cbInfo
         addrName, paramsName = self.cfgCmdGen(
             authData, transportTarget
         )
-        __varBinds = []
-        for varName, varVal in varBinds:
-            if isinstance(varName, MibVariable):
-                varName.resolveWithMib(self.mibViewController)
-                if not isinstance(varVal, base.AbstractSimpleAsn1Item):
-                    varVal = varName.getMibNode().getSyntax().clone(varVal)
-            elif isinstance(varName[0], tuple):  # legacy
-                varName = MibVariable(varName[0][0], varName[0][1], *varName[1:]).resolveWithMib(self.mibViewController)
-                if not isinstance(varVal, base.AbstractSimpleAsn1Item):
-                    varVal = varName.getMibNode().getSyntax().clone(varVal)
-            else:
-                if isinstance(varVal, base.AbstractSimpleAsn1Item):
-                    varName = MibVariable(varName).resolveWithMib(self.mibViewController, oidOnly=True)
-                else:
-                    varName = MibVariable(varName).resolveWithMib(self.mibViewController)
-                    varVal = varName.getMibNode().getSyntax().clone(varVal)
 
-            __varBinds.append((varName, varVal))
+
         return cmdgen.SetCommandGenerator().sendReq(
             self.snmpEngine,
             addrName,
-            __varBinds,
+            self.makeVarBinds(varBinds),
             __cbFun,
             (lookupNames, lookupValues, cbFun, cbCtx),
-            authData.contextEngineId,
-            authData.contextName
+            contextEngineId, contextName
         )
     
     asyncSetCmd = setCmd
     
     def nextCmd(self, authData, transportTarget, varNames, cbInfo,
-                lookupNames=False, lookupValues=False):
+                lookupNames=False, lookupValues=False,
+                contextEngineId=None, contextName=null):
         def __cbFun(sendRequestHandle,
                     errorIndication, errorStatus, errorIndex,
                     varBindTable, cbCtx):
@@ -267,6 +335,10 @@ class AsynCommandGenerator:
                 cbCtx
             )
 
+        # for backward compatibility
+        if contextName is null and authData.contextName:
+            contextName = authData.contextName
+        
         (cbFun, cbCtx) = cbInfo
         addrName, paramsName = self.cfgCmdGen(
             authData, transportTarget
@@ -277,14 +349,15 @@ class AsynCommandGenerator:
             self.makeReadVarBinds(varNames),
             __cbFun,
             (lookupNames, lookupValues, cbFun, cbCtx),
-            authData.contextEngineId, authData.contextName
+            contextEngineId, contextName
         )
 
     asyncNextCmd = nextCmd
     
     def bulkCmd(self, authData, transportTarget,
                 nonRepeaters, maxRepetitions, varNames, cbInfo,
-                lookupNames=False, lookupValues=False):
+                lookupNames=False, lookupValues=False,
+                contextEngineId=None, contextName=null):
         def __cbFun(sendRequestHandle,
                     errorIndication, errorStatus, errorIndex,
                     varBindTable, cbCtx):
@@ -298,6 +371,10 @@ class AsynCommandGenerator:
                 cbCtx
             )
 
+        # for backward compatibility
+        if contextName is null and authData.contextName:
+            contextName = authData.contextName
+        
         (cbFun, cbCtx) = cbInfo
         addrName, paramsName = self.cfgCmdGen(
             authData, transportTarget
@@ -310,7 +387,7 @@ class AsynCommandGenerator:
             varBinds,
             __cbFun,
             (lookupNames, lookupValues, cbFun, cbCtx),
-            authData.contextEngineId, authData.contextName
+            contextEngineId, contextName
         )
 
     asyncBulkCmd = bulkCmd
@@ -342,7 +419,9 @@ class CommandGenerator:
             varNames,
             (__cbFun, appReturn),
             kwargs.get('lookupNames', False),
-            kwargs.get('lookupValues', False)
+            kwargs.get('lookupValues', False),
+            kwargs.get('contextEngineId'),
+            kwargs.get('contextName', null)
         )
         self.__asynCmdGen.snmpEngine.transportDispatcher.runDispatcher()
         return (
@@ -361,9 +440,6 @@ class CommandGenerator:
             appReturn['errorIndex'] = errorIndex
             appReturn['varBinds'] = varBinds
 
-        lookupNames = kwargs.get('lookupNames', False)        
-        lookupValues = kwargs.get('lookupValues', False)
-
         appReturn = {}
         self.__asynCmdGen.setCmd(
             authData,
@@ -371,7 +447,9 @@ class CommandGenerator:
             varBinds,
             (__cbFun, appReturn),
             kwargs.get('lookupNames', False),
-            kwargs.get('lookupValues', False)
+            kwargs.get('lookupValues', False),
+            kwargs.get('contextEngineId'),
+            kwargs.get('contextName', null)
         )
         self.__asynCmdGen.snmpEngine.transportDispatcher.runDispatcher()
         return (
@@ -442,6 +520,8 @@ class CommandGenerator:
 
         lookupNames = kwargs.get('lookupNames', False)        
         lookupValues = kwargs.get('lookupValues', False)
+        contextEngineId = kwargs.get('contextEngineId')
+        contextName = kwargs.get('contextName', null)
         lexicographicMode = kwargs.get('lexicographicMode', False)
         maxRows = kwargs.get('maxRows', 0)
         ignoreNonIncreasingOid = kwargs.get('ignoreNonIncreasingOid', False)
@@ -454,7 +534,8 @@ class CommandGenerator:
             transportTarget,
             varNames,
             (__cbFun, (self, varBindHead, [], appReturn)),
-            lookupNames, lookupValues
+            lookupNames, lookupValues,
+            contextEngineId, contextName
         )
 
         self.__asynCmdGen.snmpEngine.transportDispatcher.runDispatcher()
@@ -529,6 +610,8 @@ class CommandGenerator:
 
         lookupNames = kwargs.get('lookupNames', False)        
         lookupValues = kwargs.get('lookupValues', False)
+        contextEngineId = kwargs.get('contextEngineId')
+        contextName = kwargs.get('contextName', null)
         lexicographicMode = kwargs.get('lexicographicMode', False)
         maxRows = kwargs.get('maxRows', 0)
         ignoreNonIncreasingOid = kwargs.get('ignoreNonIncreasingOid', False)
@@ -543,7 +626,8 @@ class CommandGenerator:
             nonRepeaters, maxRepetitions,
             varNames,
             (__cbFun, (self, varBindHead, [], appReturn)),
-            lookupNames, lookupValues
+            lookupNames, lookupValues,
+            contextEngineId, contextName
         )
 
         self.__asynCmdGen.snmpEngine.transportDispatcher.runDispatcher()
